# Constant values defined for grid search reduction
_UPPER_BOUND_DIFF = "upper"
_LOWER_BOUND_DIFF = "lower"
_LABEL = "y"
_EVENT = "event"
_GROUP_ID = "group_id"
_P_ATTR = "p_attr"
_ALL = "all"
_PRED = "pred"
_SIGNED = "signed"
_LOSS = "loss"
_PREDICTION = "prediction"

# Constant values defined for exponentiated gradient reductions
SHRINK_ETA = 0.8
SHRINK_REGRET = 0.8
MIN_ITER = 5
REGRET_CHECK_START_T = 5
REGRET_CHECK_INCREASE_T = 1.6
ACCURACY_MUL = 0.5
PRECISION = 1e-8
